searchNodes=[{"doc":"Documentation for Exgencode.","ref":"Exgencode.html","title":"Exgencode","type":"module"},{"doc":"This macro allows for the definition of binary PDUs in a simple way allowing for convienient encoding and decoding them between binary format and Elixir structures. PDUs Each PDU for the protocol is defined given a name that must be a valid Elixir structure (module) name followed by a list of fields that the given PDU has. Fields Each field can have the following options specified: size Defines the size of field in bits. If the field is of type :subrecord the :size is unused. defpdu SomePdu , someField : [ size : 12 ] default Defines the default value that the field should assume when building a new Elixir structure of the given PDU. defpdu PduWithDefault , aFieldWithDefault : [ size : 10 , default : 15 ] type Defines the type of the field. Field can be of type: :constant :subrecord :string :binary :float :integer :variable :skip If no type should is specified it will default to :integer . Both :integer and :float specify normal numerical values and have no special properties. :constant If the field is constant it will not become part of the Elixir structure and will not be accessible. However it will still be encoded into the binary representation and the decoding will expect the field to be present and have the given value in the decoded binary. Otherwise FunctionClauseError will be raised. A :constant field MUST have a default value specified. defpdu PduWithConstant , aConstantField : [ size : 12 , default : 10 , type : :constant ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.PduWithConstant { } ) &lt;&lt; 10 :: size ( 16 ) &gt;&gt; iex&gt; % TestPdu.PduWithConstant { } . aConstantField ** (KeyError) key :aConstantField not found in: %Exgencode.TestPdu.PduWithConstant{} :subrecord If the field is meant to contain a sub-structure then it should be of type :subrecord. Such field must have either a default value specified that is of the type of the subrecord. Alternatively it must define custom decode and encode functions. Examples: defpdu SubPdu , someField : [ size : 10 , default : 1 ] defpdu TopPdu , aField : [ size : 24 ] , subPdu : [ type : :subrecord , default : % SupPdu { } ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.TopPdu { aField : 24 } ) &lt;&lt; 24 :: size ( 24 ) , 1 :: size ( 16 ) &gt;&gt; iex&gt; Exgencode.Pdu . decode ( % TestPdu.TopPdu { } , &lt;&lt; 24 :: size ( 24 ) , 1 :: size ( 16 ) &gt;&gt; ) { % TestPdu.TopPdu { aField : 24 , subPdu : % TestPdu.SubPdu { someField : 1 } } , &lt;&lt; &gt;&gt; } :virtual The virtual fields are never encoded into binaries and exist only in the Elixir structs. When decoding into a struct the virtual field will always assume the default value. Examples: defpdu VirtualPdu , real_field : [ size : 16 ] , virtual_field : [ type : :virtual ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.VirtualPdu { real_field : 12 , virtual_field : &quot;Any value goes here&quot; } ) &lt;&lt; 12 :: size ( 16 ) &gt;&gt; iex&gt; Exgencode.Pdu . decode ( % TestPdu.VirtualPdu { } , &lt;&lt; 12 :: size ( 16 ) &gt;&gt; ) { % TestPdu.VirtualPdu { real_field : 12 , virtual_field : nil } , &lt;&lt; &gt;&gt; } :binary If the field is an arbitrary binary value it can be specified with this type. In such case the size parameter indicates size in bytes rather than bits. This type does not define any padding, that is the size of the binary that is contained in this field must be of at least the defined field size, otherwise an ArgumentError is raised. If the size is larger the binary will be trimmed. :variable Variable fields have no pre-defined size, instead the size is defined by the value of another field. When defining a :variable field, the :size parameter must be set to the name of the field definining the size, which in turn should be an :integer field. The size in that case can only be specified in bytes. All :variable fields are binary fields. Examples: defpdu VariablePdu , some_field : [ size : 16 ] , size_field : [ size : 16 ] , variable_field : [ type : :variable , size : :size_field ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.VariablePdu { some_field : 52 , size_field : 2 , variable_field : &quot;AB&quot; } ) &lt;&lt; 52 :: size ( 16 ) , 2 :: size ( 16 ) , &quot;A&quot; , &quot;B&quot; &gt;&gt; iex&gt; Exgencode.Pdu . decode ( % TestPdu.VariablePdu { } , &lt;&lt; 52 :: size ( 16 ) , 2 :: size ( 16 ) , &quot;A&quot; , &quot;B&quot; &gt;&gt; ) { % TestPdu.VariablePdu { some_field : 52 , size_field : 2 , variable_field : &quot;AB&quot; } , &lt;&lt; &gt;&gt; } Note that the field defining the size must be defined before the variable length field. Examples: defpdu BinaryMsg , someHeader : [ size : 8 , default : 10 ] , binaryField : [ size : 16 , type : :binary ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.BinaryMsg { binaryField : &quot;16charactershere&quot; } ) &lt;&lt; 10 :: size ( 8 ) , &quot;16charactershere&quot; :: binary &gt;&gt; :string The :string type is similar to :binary , however it will not raise any errors if the length of the value to be encoded is different than declared field size. Instead, the string will be trimmed if its too long and padded with trailing 0-bytes if it is too short. Upon decoded all trailing 0-bytes will be removed. For any other handling of padding or empty bytes custom decode and encode functions must be defined. Examples: defpdu StringMsg , someHeader : [ size : 8 , default : 10 ] , stringField : [ size : 16 , type : :string ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.StringMsg { stringField : &quot;16charactershere&quot; } ) &lt;&lt; 10 :: size ( 8 ) , &quot;16charactershere&quot; :: binary &gt;&gt; iex&gt; Exgencode.Pdu . encode ( % TestPdu.StringMsg { stringField : &quot;Too long string for field size&quot; } ) &lt;&lt; 10 :: size ( 8 ) , &quot;Too long string &quot; :: binary &gt;&gt; iex&gt; Exgencode.Pdu . encode ( % TestPdu.StringMsg { stringField : &quot;Too short&quot; } ) &lt;&lt; 10 :: size ( 8 ) , &quot;Too short&quot; :: binary , 0 , 0 , 0 , 0 , 0 , 0 , 0 &gt;&gt; iex&gt; Exgencode.Pdu . decode ( % TestPdu.StringMsg { } , &lt;&lt; 10 :: size ( 8 ) , &quot;Too short&quot; :: binary , 0 , 0 , 0 , 0 , 0 , 0 , 0 &gt;&gt; ) { % TestPdu.StringMsg { stringField : &quot;Too short&quot; } , &lt;&lt; &gt;&gt; } :skip If the field will never be decoded into the struct. It's value will be set to :default when encoding the struct. defpdu SkippedPdu , testField : [ default : 1 , size : 16 ] , skippedField : [ size : 8 , default : 5 , type : :skip ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.SkippedPdu { testField : 15 } ) &lt;&lt; 15 :: size ( 16 ) , 5 :: size ( 8 ) &gt;&gt; iex&gt; % TestPdu.SkippedPdu { } . skippedField ** (KeyError) key :skippedField not found in: %Exgencode.TestPdu.SkippedPdu{testField: 1} iex&gt; Exgencode.Pdu . decode ( % TestPdu.SkippedPdu { } , &lt;&lt; 32 :: size ( 16 ) , 11 :: size ( 8 ) &gt;&gt; ) { % TestPdu.SkippedPdu { testField : 32 } , &lt;&lt; &gt;&gt; } encode/decode Defines a custom encode or decode function. See type specifications for the function specification. If a PDU has a custom encode function defined it must also define a custom decode function. Custom encode and decode functions can override any of the other parameters the field has if the user wishes it so. Examples: defpdu CustomPdu , normalField : [ size : 16 , default : 3 ] , customField : [ encode : fn ( val ) -&gt; &lt;&lt; val * 2 :: size ( 12 ) &gt;&gt; end , decode : fn ( pdu , &lt;&lt; val :: size ( 12 ) &gt;&gt; ) -&gt; { struct ( pdu , %{ customField : div ( val , 2 ) } ) , &lt;&lt; &gt;&gt; } end ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.CustomPdu { customField : 10 } ) &lt;&lt; 3 :: size ( 16 ) , 20 :: size ( 16 ) &gt;&gt; iex&gt; Exgencode.Pdu . decode ( % TestPdu.CustomPdu { } , &lt;&lt; 3 :: size ( 16 ) , 20 :: size ( 16 ) &gt;&gt; ) { % TestPdu.CustomPdu { customField : 10 } , &lt;&lt; &gt;&gt; } version Defines the requirement for the protocol version for the given field to be included in the message. When a version is specified encode/2 and decode/3 can take an optional parameter with the given version name. If the given version matches the version requirement defined by this option in the PDU definition, the field will be included. Otherwise it will be skipped. defpdu VersionedMsg , oldField : [ default : 10 , size : 16 ] , newerField : [ size : 8 , version : &quot;&gt;= 2.0.0&quot; ] , See documentation for Exgencode.Pdu./2 for examples. endianness Defines the endianness of the particular field. Allowed options are :big , :little and :native . Defaults to :big Examples: defpdu EndianMsg , bigField : [ default : 15 , size : 32 , endianness : :big ] , smallField : [ default : 15 , size : 32 , endianness : :little ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.EndianMsg { } ) &lt;&lt; 15 :: big - size ( 32 ) , 15 :: little - size ( 32 ) &gt;&gt; conditional Defines that the field is present in encoded binary format only if another field has a non-null value. Examples: defpdu ConditionalPdu , normal_field : [ size : 16 ] , flag_field : [ size : 8 ] , conditional_field : [ size : 8 , conditional : :flag_field ] , another_normal_field : [ size : 8 ] , second_flag : [ size : 8 ] , size_field : [ size : 16 , conditional : :second_flag ] , conditional_variable_field : [ type : :variable , size : :size_field , conditional : :second_flag ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.ConditionalPdu { ...&gt; normal_field : 12 , ...&gt; flag_field : 1 , ...&gt; conditional_field : 10 , ...&gt; another_normal_field : 200 , ...&gt; second_flag : 1 , ...&gt; size_field : 4 , ...&gt; conditional_variable_field : &quot;test&quot; ...&gt; } ) &lt;&lt; 12 :: size ( 16 ) , 1 , 10 , 200 , 1 , 4 :: size ( 16 ) , &quot;test&quot; &gt;&gt; iex&gt; Exgencode.Pdu . encode ( % TestPdu.ConditionalPdu { ...&gt; normal_field : 12 , ...&gt; flag_field : 1 , ...&gt; conditional_field : 10 , ...&gt; another_normal_field : 200 , ...&gt; second_flag : 0 , ...&gt; size_field : nil , ...&gt; conditional_variable_field : nil ...&gt; } ) &lt;&lt; 12 :: size ( 16 ) , 1 , 10 , 200 , 0 &gt;&gt; offset_to Defines that the field contains the offset to another field. The offset is in bytes since the beginning of the PDU. Note that offsets are automatically calculated when calling Exgencode.Pdu.encode/2 Examples: defpdu OffsetPdu , offset_to_field_a : [ size : 16 , offset_to : :field_a ] , offset_to_field_b : [ size : 16 , offset_to : :field_b ] , offset_to_field_c : [ size : 16 , offset_to : :field_c ] , field_a : [ size : 8 ] , size_field : [ size : 16 ] , variable_field : [ type : :variable , size : :size_field ] , field_b : [ size : 8 ] , field_c : [ size : 8 , conditional : :offset_to_field_c ] iex&gt; Exgencode.Pdu . encode ( % TestPdu.OffsetPdu { ...&gt; field_a : 14 , ...&gt; size_field : 4 , ...&gt; variable_field : &quot;test&quot; , ...&gt; field_b : 15 , ...&gt; field_c : 20 ...&gt; } ) &lt;&lt; 6 :: size ( 16 ) , 9 + 4 :: size ( 16 ) , 10 + 4 :: size ( 16 ) , 14 , 4 :: size ( 16 ) &gt;&gt; &lt;&gt; &quot;test&quot; &lt;&gt; &lt;&lt; 15 , 20 &gt;&gt; iex&gt; Exgencode.Pdu . encode ( % TestPdu.OffsetPdu { ...&gt; field_a : 14 , ...&gt; size_field : 4 , ...&gt; variable_field : &quot;test&quot; , ...&gt; field_b : 15 , ...&gt; field_c : nil ...&gt; } ) &lt;&lt; 6 :: size ( 16 ) , 9 + 4 :: size ( 16 ) , 0 :: size ( 16 ) , 14 , 4 :: size ( 16 ) &gt;&gt; &lt;&gt; &quot;test&quot; &lt;&gt; &lt;&lt; 15 &gt;&gt;","ref":"Exgencode.html#defpdu/2","title":"Exgencode.defpdu/2","type":"macro"},{"doc":"Parameters of the given field","ref":"Exgencode.html#t:fieldParam/0","title":"Exgencode.fieldParam/0","type":"type"},{"doc":"A custom decoding function that receives the PDU decoded so far and remaining binary and is meant to return PDU with the field decoded and remaining binary.","ref":"Exgencode.html#t:field_decode_fun/0","title":"Exgencode.field_decode_fun/0","type":"type"},{"doc":"A custom encoding function that is meant to take the value of the field and return its binary represantion.","ref":"Exgencode.html#t:field_encode_fun/0","title":"Exgencode.field_encode_fun/0","type":"type"},{"doc":"The endianness the field should be encoded/decoded with","ref":"Exgencode.html#t:field_endianness/0","title":"Exgencode.field_endianness/0","type":"type"},{"doc":"Name of the field.","ref":"Exgencode.html#t:field_name/0","title":"Exgencode.field_name/0","type":"type"},{"doc":"The type of the field.","ref":"Exgencode.html#t:field_type/0","title":"Exgencode.field_type/0","type":"type"},{"doc":"A PDU, that is an Elixir structure representing a PDU.","ref":"Exgencode.html#t:pdu/0","title":"Exgencode.pdu/0","type":"type"},{"doc":"PDU name, must be a structure name","ref":"Exgencode.html#t:pdu_name/0","title":"Exgencode.pdu_name/0","type":"type"},{"doc":"Desired return type of pdu size","ref":"Exgencode.html#t:return_size_type/0","title":"Exgencode.return_size_type/0","type":"type"},{"doc":"Helper functions for generating encoding and decoding functions.","ref":"Exgencode.EncodeDecode.html","title":"Exgencode.EncodeDecode","type":"module"},{"doc":"","ref":"Exgencode.EncodeDecode.html#create_decode_fun/3","title":"Exgencode.EncodeDecode.create_decode_fun/3","type":"function"},{"doc":"","ref":"Exgencode.EncodeDecode.html#create_encode_fun/3","title":"Exgencode.EncodeDecode.create_encode_fun/3","type":"function"},{"doc":"","ref":"Exgencode.EncodeDecode.html#create_versioned_decode/2","title":"Exgencode.EncodeDecode.create_versioned_decode/2","type":"function"},{"doc":"","ref":"Exgencode.EncodeDecode.html#create_versioned_encode/2","title":"Exgencode.EncodeDecode.create_versioned_encode/2","type":"function"},{"doc":"","ref":"Exgencode.EncodeDecode.html#wrap_custom_encode/2","title":"Exgencode.EncodeDecode.wrap_custom_encode/2","type":"function"},{"doc":"Helper functions for generating offset calculating functions.","ref":"Exgencode.Offsets.html","title":"Exgencode.Offsets","type":"module"},{"doc":"","ref":"Exgencode.Offsets.html#create_offset_fun/1","title":"Exgencode.Offsets.create_offset_fun/1","type":"function"},{"doc":"","ref":"Exgencode.Offsets.html#create_offset_fun/3","title":"Exgencode.Offsets.create_offset_fun/3","type":"function"},{"doc":"","ref":"Exgencode.Offsets.html#offset_dependencies/2","title":"Exgencode.Offsets.offset_dependencies/2","type":"function"},{"doc":"The module contains functions for operating with PDUs defined with the defpdu/2 macro.","ref":"Exgencode.Pdu.html","title":"Exgencode.Pdu","type":"module"},{"doc":"Decode a binary into the specified Elixir structure. Returns the given structure with fields filled out and the remainder binary. The remainder should be an empty binary and leftovers usually indicate a mangled binary. Examples: iex&gt; Exgencode.Pdu . decode ( % TestPdu.PzTestMsg { } , &lt;&lt; 1 :: size ( 12 ) , 100 :: size ( 24 ) , 15 :: size ( 8 ) , 10 :: size ( 28 ) &gt;&gt; ) { % TestPdu.PzTestMsg { otherTestField : 100 } , &lt;&lt; &gt;&gt; } Version number can be optionally added to control how the decoding function reads the given binary. If the provided version does not match the requirement specified in the field definition the given field will be ignored. Examples: iex&gt; Exgencode.Pdu . decode ( % TestPdu.VersionedMsg { } , &lt;&lt; 10 :: size ( 16 ) &gt;&gt; , &quot;1.0.0&quot; ) { % TestPdu.VersionedMsg { oldField : 10 } , &lt;&lt; &gt;&gt; } iex&gt; Exgencode.Pdu . decode ( % TestPdu.VersionedMsg { } , &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) &gt;&gt; , &quot;2.0.0&quot; ) { % TestPdu.VersionedMsg { oldField : 10 , newerField : 111 } , &lt;&lt; &gt;&gt; }","ref":"Exgencode.Pdu.html#decode/3","title":"Exgencode.Pdu.decode/3","type":"function"},{"doc":"Encode the Elixir structure into a binary give the protocol version. Examples: iex&gt; Exgencode.Pdu . encode ( % TestPdu.PzTestMsg { otherTestField : 100 } ) &lt;&lt; 1 :: size ( 12 ) , 100 :: size ( 24 ) , 15 :: size ( 8 ) , 10 :: size ( 28 ) &gt;&gt; iex&gt; Exgencode.Pdu . encode ( % TestPdu.PzTestMsg { testField : 99 , otherTestField : 100 } ) &lt;&lt; 99 :: size ( 12 ) , 100 :: size ( 24 ) , 15 :: size ( 8 ) , 10 :: size ( 28 ) &gt;&gt; Version number can be optionally added to control the encoding of the PDU and exclude certain fields if the version number is lower that specified. pdu = % TestPdu.VersionedMsg { newerField : 111 , evenNewerField : 7 } assert &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) , 14 :: size ( 8 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu ) assert &lt;&lt; 10 :: size ( 16 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu , &quot;1.0.0&quot; ) assert &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu , &quot;2.0.0&quot; ) Examples: iex&gt; Exgencode.Pdu . encode ( % TestPdu.VersionedMsg { newerField : 111 , evenNewerField : 7 } , &quot;1.0.0&quot; ) &lt;&lt; 10 :: size ( 16 ) &gt;&gt; iex&gt; Exgencode.Pdu . encode ( % TestPdu.VersionedMsg { newerField : 111 , evenNewerField : 7 } , &quot;2.0.0&quot; ) &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) &gt;&gt;","ref":"Exgencode.Pdu.html#encode/2","title":"Exgencode.Pdu.encode/2","type":"function"},{"doc":"Set the values of all offset fields. If no fields have the :offset_to property it become an identity function","ref":"Exgencode.Pdu.html#set_offsets/2","title":"Exgencode.Pdu.set_offsets/2","type":"function"},{"doc":"Returns the size of the field in bits.","ref":"Exgencode.Pdu.html#sizeof/2","title":"Exgencode.Pdu.sizeof/2","type":"function"},{"doc":"Returns the size of the given version pdu. In case the version argument is not passed it returns full size (all fields defined within defpdu macro). Does not count subrecords size .","ref":"Exgencode.Pdu.html#sizeof_pdu/3","title":"Exgencode.Pdu.sizeof_pdu/3","type":"function"},{"doc":"","ref":"Exgencode.Pdu.Protocol.html","title":"Exgencode.Pdu.Protocol","type":"protocol"},{"doc":"Decode a binary into the specified Elixir structure.","ref":"Exgencode.Pdu.Protocol.html#decode/3","title":"Exgencode.Pdu.Protocol.decode/3","type":"function"},{"doc":"Encode the Elixir structure into a binary given the protocol version.","ref":"Exgencode.Pdu.Protocol.html#encode/2","title":"Exgencode.Pdu.Protocol.encode/2","type":"function"},{"doc":"Calculate the values of all offset fields.","ref":"Exgencode.Pdu.Protocol.html#set_offsets/2","title":"Exgencode.Pdu.Protocol.set_offsets/2","type":"function"},{"doc":"Returns the size of the field in bits.","ref":"Exgencode.Pdu.Protocol.html#sizeof/2","title":"Exgencode.Pdu.Protocol.sizeof/2","type":"function"},{"doc":"Returns the size of the pdu for given version.","ref":"Exgencode.Pdu.Protocol.html#sizeof_pdu/3","title":"Exgencode.Pdu.Protocol.sizeof_pdu/3","type":"function"},{"doc":"","ref":"Exgencode.Pdu.Protocol.html#t:t/0","title":"Exgencode.Pdu.Protocol.t/0","type":"type"},{"doc":"Helper functions for generating sizeof/2 protocol function implementation.","ref":"Exgencode.Sizeof.html","title":"Exgencode.Sizeof","type":"module"},{"doc":"","ref":"Exgencode.Sizeof.html#build_sizeof/1","title":"Exgencode.Sizeof.build_sizeof/1","type":"function"},{"doc":"","ref":"Exgencode.Sizeof.html#build_sizeof_pdu/1","title":"Exgencode.Sizeof.build_sizeof_pdu/1","type":"function"},{"doc":"Helper functions for field validation","ref":"Exgencode.Validator.html","title":"Exgencode.Validator","type":"module"},{"doc":"","ref":"Exgencode.Validator.html#validate_field/4","title":"Exgencode.Validator.validate_field/4","type":"function"},{"doc":"","ref":"Exgencode.Validator.html#validate_pdu/2","title":"Exgencode.Validator.validate_pdu/2","type":"function"},{"doc":"Exgencode","ref":"readme.html","title":"Exgencode","type":"extras"},{"doc":"This package allows for simple definition of binary-based protocols together with an Elixir protocol that allows for simple transforming of binary packages into Elixir structures and vice versa. The protocols are defined by creating a module for each protocole and utilising the PDU definitions to define all messages the protocol needs. See the following sections for more.","ref":"readme.html#description","title":"Exgencode - Description","type":"extras"},{"doc":"The package is available in Hex and can be installed by adding exgencode to your list of dependencies in mix.exs : def deps do [ { :exgencode , &quot;~&gt; 1.0.0&quot; } ] end Documentation can be generated with ExDoc . The documentation can also be found at https://hexdocs.pm/exgencode .","ref":"readme.html#installation","title":"Exgencode - Installation","type":"extras"},{"doc":"Protocol messages can be defined using the defpdu macro. Each message is defined with its name and a list of message fields and their options. Each field should define the size (given in bits) for the particular and can define a default value. For example: defpdu ExamplePdu , firstField : [ size : 10 ] , fieldWithADefault : [ size : 14 , default : 5 ] Once defined the code can use the messages as structure e.g. %ExamplePdu{firstField: 5} Messages can also be nested within other messages. In that case the field must by of type :subrecord and must define the default which indicates which what type of sub-structure should the field contain. defpdu SubPdu , someField : [ size : 8 , default : 1 ] defpdu MainPdu , section : [ size : 16 ] subSection : [ type : :subrecord , default : % SubPdu { } ] Alternatively each field may define custom encoding and decoding functions: defpdu CustomPdu , normalField : [ size : 16 , default : 3 ] customField : [ encode : fn ( val ) -&gt; &lt;&lt; val :: size ( 16 ) &gt;&gt; end , decode : fn ( pdu , &lt;&lt; val :: size ( 16 ) &gt;&gt; ) -&gt; { struct ( pdu , :customField =&gt; val ) , &lt;&lt; &gt;&gt; } end ] With custom functions other parameters can be omitted. It is reccommended to place all messages for a given interface in one module. defmodule MyProtocol do defpdu HelloMsg , userId : [ size : 256 ] defpdu ByeMsg , userId : [ size : 256 ] end Please note that the total size of your PDU must be in full bytes, that is its total size in bits must be divisible by 8. Pdu Protocol exgencode also provides the Exgencode.Pdu.Protocol protocol that each pdu defined with defpdu will automatically implement. The Exgencode.Pdu module can be used to transform between binary and structures. defpdu MsgSubSection , someField : [ default : 15 , size : 8 ] defpdu PzTestMsg , testField : [ default : 1 , size : 12 ] , otherTestField : [ size : 24 ] , subSection : [ default : % MsgSubSection { } , type : :subrecord ] , constField : [ default : 10 , size : 24 , type : :constant ] test &quot;encoding&quot; do pdu = % TestPdu.PzTestMsg { otherTestField : 100 } assert &lt;&lt; 1 :: size ( 12 ) , 100 :: size ( 24 ) , 15 :: size ( 8 ) , 10 :: size ( 24 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu ) end test &quot;decoding&quot; do pdu = % TestPdu.PzTestMsg { otherTestField : 100 } binary = &lt;&lt; 1 :: size ( 12 ) , 100 :: size ( 24 ) , 15 :: size ( 8 ) , 10 :: size ( 24 ) &gt;&gt; assert { ^ pdu , &lt;&lt; &gt;&gt; } = Exgencode.Pdu . decode ( % TestPdu.PzTestMsg { } , binary ) end Versioning The fields can also be versioned allowing for specifications of versions of the protocol defined by the defpdu blocks. Both encode/2 and decode/3 can take the version number that are meant to be used to encode or decode. This allows the protocol to easily operate in backwards compatibility. test &quot;versioning encode&quot; do pdu = % TestPdu.VersionedMsg { newerField : 111 , evenNewerField : 7 } assert &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) , 14 :: size ( 8 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu ) assert &lt;&lt; 10 :: size ( 16 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu , &quot;1.0.0&quot; ) assert &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu , &quot;2.0.0&quot; ) assert &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) , 14 :: size ( 8 ) &gt;&gt; == Exgencode.Pdu . encode ( pdu , &quot;2.1.0&quot; ) end test &quot;versioning decode&quot; do pdu = % TestPdu.VersionedMsg { } binary = &lt;&lt; 10 :: size ( 16 ) &gt;&gt; assert { ^ pdu , &lt;&lt; &gt;&gt; } = Exgencode.Pdu . decode ( % TestPdu.VersionedMsg { } , binary , &quot;1.0.0&quot; ) pdu = % TestPdu.VersionedMsg { newerField : 111 } binary = &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) &gt;&gt; assert { ^ pdu , &lt;&lt; &gt;&gt; } = Exgencode.Pdu . decode ( % TestPdu.VersionedMsg { } , binary , &quot;2.0.0&quot; ) pdu = % TestPdu.VersionedMsg { newerField : 111 , evenNewerField : 7 } binary = &lt;&lt; 10 :: size ( 16 ) , 111 :: size ( 8 ) , 14 :: size ( 8 ) &gt;&gt; assert { ^ pdu , &lt;&lt; &gt;&gt; } = Exgencode.Pdu . decode ( % TestPdu.VersionedMsg { } , binary ) assert { ^ pdu , &lt;&lt; &gt;&gt; } = Exgencode.Pdu . decode ( % TestPdu.VersionedMsg { } , binary , &quot;2.1.0&quot; ) assert { % TestPdu.VersionedMsg { } , &lt;&lt; 111 :: size ( 8 ) , 14 :: size ( 8 ) &gt;&gt; } = Exgencode.Pdu . decode ( % TestPdu.VersionedMsg { } , binary , &quot;1.0.0&quot; ) end See the documentation for details.","ref":"readme.html#usage","title":"Exgencode - Usage","type":"extras"}]